import itertools
import typing
from enum import Enum
from uuid import UUID

from .auxdata import AuxData, AuxDataContainer
from .block import ByteBlock, CfgNode, CodeBlock, DataBlock, ProxyBlock
from .byteinterval import ByteInterval, SymbolicExpressionElement
from .node import Node
from .proto import Module_pb2
from .section import Section
from .symbol import Symbol
from .util import (
    DictLike,
    SetWrapper,
    nodes_at,
    nodes_on,
    symbolic_expressions_at,
)


class Module(AuxDataContainer):
    """Represents a loadable object, such as an executable or library.

    :ivar ~.binary_path: The path to the loadable binary object
        represented by this module. An empty string if not specified.
        The file represented by this path is indicitave of what file
        this ``Module`` was initially created from; it is not guaranteed to
        currently exist or have the same contents.
    :ivar ~.isa: The ISA of the binary.
    :ivar ~.file_format: The file format of the binary.
    :ivar ~.name: The name given to the binary. Some file formats use this
        for linking and/or symbol resolution purposes. An empty string if
        not specified by the format.
    :ivar ~.preferred_addr: The preferred loading address of the binary.
    :ivar ~.proxies: A set containing all the :class:`gtirb.ProxyBlock`\\s
        in the binary.
    :ivar ~.rebase_delta: The rebase delta of the binary.
    :ivar ~.sections: A set containing all the :class:`gtirb.Section`\\s
        in the binary.
    :ivar ~.symbols: A set containing all the :class:`gtirb.Symbol`\\s
        in the binary.
    :ivar ~.entry_point: A :class:`CodeBlock` representing where
        control flow of this module begins at, or None if not present.
    """

    class FileFormat(Enum):
        """Identifies the executable file format of the binary represented
        by a :class:`gtirb.Module`.
        """

        Undefined = Module_pb2.FileFormat.Value("Format_Undefined")
        """A file format that has not yet been specified.
        This is for unitialized modules; do not use to refer to
        file formats without ``FileFormat`` values.
        """

        COFF = Module_pb2.FileFormat.Value("COFF")
        """The Common Object File Format."""

        ELF = Module_pb2.FileFormat.Value("ELF")
        """The Executable and Linkable Format,
        formerly the Extensible Linking Format.
        """

        IdaProDb32 = Module_pb2.FileFormat.Value("IdaProDb32")
        """A 32-bit IDA Pro database file."""

        IdaProDb64 = Module_pb2.FileFormat.Value("IdaProDb64")
        """A 64-bit IDA Pro database file."""

        MACHO = Module_pb2.FileFormat.Value("MACHO")
        """A Mach object file."""

        PE = Module_pb2.FileFormat.Value("PE")
        """Microsoft's Portable Executable format."""

        RAW = Module_pb2.FileFormat.Value("RAW")
        """A raw binary file, with no file format."""

        XCOFF = Module_pb2.FileFormat.Value("XCOFF")
        """The Extended Common Object File Format."""

    class ISA(Enum):
        """Identifies the instruction set architecture (ISA)
        targeted by a :class:`gtirb.Module`.
        """

        Undefined = Module_pb2.ISA.Value("ISA_Undefined")
        """An ISA that has not yet been specified.
        This is for unitialized modules;
        use :class:`gtirb.Module.ISA.ValidButUnsupported`
        instead for specifying undefined ISAs.
        """

        ARM = Module_pb2.ISA.Value("ARM")
        """The Acorn RISC Machine, 32-bit."""

        ARM64 = Module_pb2.ISA.Value("ARM64")
        """The Acorn RISC Machine, 64-bit."""

        IA32 = Module_pb2.ISA.Value("IA32")
        """The 32-bit Intel Architecture. Also known as i386, x86, or x32."""

        PPC32 = Module_pb2.ISA.Value("PPC32")
        """IBM's 32-bit PowerPC (Performance Optimization with Enhanced RISC /
        Performance Computing) architecture."""

        PPC64 = Module_pb2.ISA.Value("PPC64")
        """IBM's 64-bit PowerPC (Performance Optimization with Enhanced RISC /
        Performance Computing) architecture."""

        MIPS32 = Module_pb2.ISA.Value("MIPS32")
        """Microprocessor without Interlocked Pipelined Stages, 32-bit."""

        MIPS64 = Module_pb2.ISA.Value("MIPS64")
        """Microprocessor without Interlocked Pipelined Stages, 64-bit."""

        X64 = Module_pb2.ISA.Value("X64")
        """The 64-bit Intel Architecture. Also known as x86_64."""

        ValidButUnsupported = Module_pb2.ISA.Value("ValidButUnsupported")
        """An unknown or undefined ISA."""

    class _NodeSet(SetWrapper):
        def __init__(self, node, field, *args):
            self._node = node  # type: Module
            self._field = field  # type: str
            super().__init__(*args)

        def add(self, v):
            if v._module is not None:
                getattr(v._module, self._field).discard(v)
            v._module = self._node
            if self._node.ir is not None:
                v._add_to_uuid_cache(self._node.ir._local_uuid_cache)
            return super().add(v)

        def discard(self, v):
            if v not in self:
                return
            v._module = None
            if self._node.ir is not None:
                v._remove_from_uuid_cache(self._node.ir._local_uuid_cache)
            return super().discard(v)

    def __init__(
        self,
        *,
        aux_data=dict(),  # type: DictLike[str, AuxData]
        binary_path="",  # type: str
        file_format=FileFormat.Undefined,  # type: Module.FileFormat
        isa=ISA.Undefined,  # type: Module.ISA
        name="",  # type: str
        preferred_addr=0,  # type: int
        proxies=set(),  # type: typing.Iterable[ProxyBlock]
        rebase_delta=0,  # type: int
        sections=set(),  # type: typing.Iterable[Section]
        symbols=set(),  # type: typing.Iterable[Symbol]
        entry_point=None,  # type: typing.Optional[CodeBlock]
        uuid=None  # type: typing.Optional[UUID]
    ):
        # type: (...) -> None
        """
        :param aux_data: The initial auxiliary data to be associated
            with the object, as a mapping from names to
            :class:`gtirb.AuxData`, defaults to an empty :class:`dict`.
        :param binary_path: The path to the loadable binary object
            represented by this module.
        :param isa: The ISA of the binary.
        :param file_format: The file format of the binary.
        :param name: The name given to the binary.
        :param preferred_addr: The preferred loading address of the binary.
        :param proxies: A set containing all the :class:`gtirb.ProxyBlock`\\s
            in the binary.
        :param rebase_delta: The rebase delta of the binary.
        :param sections: A set containing all the :class:`gtirb.Section`\\s
            in the binary.
        :param symbols: A set containing all the :class:`gtirb.Symbol`\\s
            in the binary.
        :param entry_point: A :class:`CodeBlock` representing where
            control flow of this module begins at, or None if not present.
        :param uuid: The UUID of this ``Module``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """

        self._ir = None  # type: "IR"
        self.binary_path = binary_path  # type: str
        self.isa = isa  # type: Module.ISA
        self.file_format = file_format  # type: Module.FileFormat
        self.name = name  # type: str
        self.preferred_addr = preferred_addr  # type: int
        self.proxies = Module._NodeSet(
            self, "proxies", proxies
        )  # type: typing.Set[ProxyBlock]
        self.rebase_delta = rebase_delta  # type: int
        self.sections = Module._NodeSet(
            self, "sections", sections
        )  # type: typing.Set[Section]
        self.symbols = Module._NodeSet(
            self, "symbols", symbols
        )  # type: typing.Set[Symbol]
        self.entry_point = entry_point  # type: typing.Optional[CodeBlock]
        # Initialize the aux data last so that the cache is populated
        super().__init__(aux_data, uuid)

    @classmethod
    def _decode_protobuf(cls, proto_module, uuid, ir):
        # type: (Module_pb2.Module. UUID, typing.Optiona["IR"]) -> Module
        m = cls(
            binary_path=proto_module.binary_path,
            isa=Module.ISA(proto_module.isa),
            file_format=Module.FileFormat(proto_module.file_format),
            name=proto_module.name,
            preferred_addr=proto_module.preferred_addr,
            rebase_delta=proto_module.rebase_delta,
            uuid=uuid,
        )
        m._add_to_uuid_cache(ir._local_uuid_cache)

        # proxies depend on nothing
        m.proxies.update(
            ProxyBlock._from_protobuf(p, ir) for p in proto_module.proxies
        )
        # sections depend on symbolic expressions, so that step is split out
        # from _decode_protobuf into _decode_symbolic_expressions
        m.sections.update(
            Section._from_protobuf(s, ir) for s in proto_module.sections
        )
        # entry point is a code block, which depends on sections
        m.entry_point = (
            ir.get_by_uuid(UUID(bytes=proto_module.entry_point))
            if proto_module.entry_point
            else None
        )
        # symbols depend on blocks
        m.symbols.update(
            Symbol._from_protobuf(s, ir) for s in proto_module.symbols
        )
        # symbolic expressions depend on symbols
        for section in m.sections:
            for interval in section.byte_intervals:
                interval._decode_symbolic_expressions(ir)
        # aux data may depend on any node
        m.aux_data.update(
            AuxDataContainer._read_protobuf_aux_data(proto_module, ir)
        )

        return m

    def _to_protobuf(self):
        # type: () -> Module_pb2.Module
        proto_module = Module_pb2.Module()
        self._write_protobuf_aux_data(proto_module)
        proto_module.binary_path = self.binary_path
        proto_module.isa = self.isa.value
        proto_module.file_format = self.file_format.value
        proto_module.name = self.name
        proto_module.preferred_addr = self.preferred_addr
        proto_module.proxies.extend(p._to_protobuf() for p in self.proxies)
        proto_module.rebase_delta = self.rebase_delta
        proto_module.sections.extend(s._to_protobuf() for s in self.sections)
        proto_module.symbols.extend(s._to_protobuf() for s in self.symbols)
        if self.entry_point is not None:
            proto_module.entry_point = self.entry_point.uuid.bytes
        proto_module.uuid = self.uuid.bytes
        return proto_module

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not super().deep_eq(other):
            return False
        if not isinstance(other, Module):
            return False
        for attr in (
            "binary_path",
            "isa",
            "file_format",
            "name",
            "preferred_addr",
            "rebase_delta",
        ):
            if getattr(self, attr) != getattr(other, attr):
                return False

        for attr in ("proxies", "sections", "symbols"):
            self_nodes = sorted(getattr(self, attr), key=lambda n: n.uuid)
            other_nodes = sorted(getattr(other, attr), key=lambda n: n.uuid)
            if not len(self_nodes) == len(other_nodes):
                return False
            for self_node, other_node in zip(self_nodes, other_nodes):
                if not self_node.deep_eq(other_node):
                    return False

        if self.entry_point is None:
            if other.entry_point is not None:
                return False
        else:
            if not self.entry_point.deep_eq(other.entry_point):
                return False

        return True

    def __repr__(self):
        # type: () -> str
        return (
            "Module("
            "uuid={uuid!r}, "
            "name={name!r}, "
            "binary_path={binary_path!r}, "
            "isa=Module.{isa!s}, "
            "file_format=Module.{file_format!s}, "
            "preferred_addr={preferred_addr:#x}, "
            "rebase_delta={rebase_delta:#x}, "
            "proxies={proxies!r}, "
            "sections={sections!r}, "
            "symbols={symbols!r}, "
            "entry_point={entry_point!r}, "
            ")".format(**self.__dict__)
        )

    @property
    def ir(self):
        # type: () -> typing.Optional["IR"]
        """The :class:`IR` this module belongs to."""

        return self._ir

    @ir.setter
    def ir(self, value):
        # type: (typing.Optional["IR"]) -> None
        if self._ir is not None:
            self._ir.modules.remove(self)
        if value is not None:
            value.modules.append(self)

    @property
    def address(self):
        # type: () -> typing.Optional[int]
        """Get the address of this module, if known.

        The address is calculated from the :class:`Section` objects in
        this module. More specifically, if the address of all sections
        in this module are fixed, then it will return the address of the
        section lowest in memory. If any one section does not have an address,
        then this will be ``None``, as the address is not calculable in that
        case. Note that a module with no sections in it has no address or
        size, so it will be ``None`` in that case.
        """

        lowest = None
        for s in self.sections:
            if s.address is None:
                return None
            if lowest is None or s.address < lowest:
                lowest = s.address
        return lowest

    @property
    def size(self):
        # type: () -> typing.Optional[int]
        """Get the size of this module, if known.

        The address is calculated from the :class:`Section` objects in
        this module. More specifically, if the address of all sections
        in this module are fixed, then it will return the difference between
        the lowest and highest address among the sections. If any one section
        does not have an address, then this will be ``None``, as the size is
        not calculable in that case. Note that a module with no sections in
        it has no address or size, so it will be ``None`` in that case.
        """

        lowest = None
        highest = None
        for s in self.sections:
            if s.address is None:
                return None
            if lowest is None or s.address < lowest:
                lowest = s.address
            if highest is None or s.address + s.size > highest:
                highest = s.address + s.size
        if lowest is None or highest is None:
            return None
        return highest - lowest

    @property
    def byte_intervals(self):
        # type: () -> typing.Iterator[ByteInterval]
        """The :class:`ByteInterval`\\s in this module."""

        return itertools.chain.from_iterable(
            s.byte_intervals for s in self.sections
        )

    @property
    def byte_blocks(self):
        # type: () -> typing.Iterator[ByteBlock]
        """The :class:`ByteBlock`\\s in this module."""

        return itertools.chain.from_iterable(
            s.byte_blocks for s in self.sections
        )

    @property
    def code_blocks(self):
        # type: () -> typing.Iterator[CodeBlock]
        """The :class:`CodeBlock`\\s in this module."""

        return itertools.chain.from_iterable(
            s.code_blocks for s in self.sections
        )

    @property
    def data_blocks(self):
        # type: () -> typing.Iterator[DataBlock]
        """The :class:`DataBlock`\\s in this module."""

        return itertools.chain.from_iterable(
            s.data_blocks for s in self.sections
        )

    @property
    def cfg_nodes(self):
        # type: () -> typing.Iterator[CfgNode]
        """The :class:`CfgNode`\\s in this module."""

        return itertools.chain(self.code_blocks, self.proxies)

    def sections_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[Section]
        """Finds all the sections that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.sections, addrs)

    def sections_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[Section]
        """Finds all the sections that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.sections, addrs)

    def byte_intervals_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
        """Finds all the byte intervals that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.byte_intervals, addrs)

    def byte_intervals_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
        """Finds all the byte intervals that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.byte_intervals, addrs)

    def byte_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
        """Finds all the byte blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.byte_blocks, addrs)

    def byte_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
        """Finds all the byte blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.byte_blocks, addrs)

    def code_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
        """Finds all the code blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.code_blocks, addrs)

    def code_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
        """Finds all the code blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.code_blocks, addrs)

    def data_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
        """Finds all the data blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.data_blocks, addrs)

    def data_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
        """Finds all the data blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.data_blocks, addrs)

    def symbolic_expressions_at(
        self, addrs  # type: typing.Union[int, range]
    ):
        # type: (...) -> typing.Iterable[SymbolicExpressionElement]
        """Finds all the symbolic expressions that begin at an address or
        range of addresses.

        :param addrs: Either a ``range`` object or a single address.
        :returns: Yields ``(interval, offset, symexpr)`` tuples for every
            symbolic expression in the range.
        """

        return symbolic_expressions_at(self.sections, addrs)

    def _add_to_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is added."""

        cache[self.uuid] = self
        for proxy in self.proxies:
            proxy._add_to_uuid_cache(cache)
        for section in self.sections:
            section._add_to_uuid_cache(cache)
        for symbol in self.symbols:
            symbol._add_to_uuid_cache(cache)

    def _remove_from_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is removed."""

        del cache[self.uuid]
        for proxy in self.proxies:
            proxy._remove_from_uuid_cache(cache)
        for section in self.sections:
            section._remove_from_uuid_cache(cache)
        for symbol in self.symbols:
            symbol._remove_from_uuid_cache(cache)



import typing
from uuid import UUID, uuid4


class Node:
    """A Node is any GTIRB object which can be referenced by UUID.

    :ivar ~.uuid: The UUID of this Node.
    """

    def __init__(self, uuid=None):
        # type: (typing.Optional[UUID]) -> None
        """
        :param uuid: The UUID of this ``Node``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """

        if uuid is None:
            uuid = uuid4()
        self.uuid = uuid  # type: UUID

    @classmethod
    def _decode_protobuf(cls, proto_object, uuid, ir):
        # type: (typing.Any, UUID, typing.Optional["IR"]) -> Node
        """Decode a Protobuf object to a Python GTIRB object.
        Must be overridden by subclasses.

        :param proto_object: The Protobuf object.
        :param uuid: The UUID of the object.
        """

        raise NotImplementedError

    @classmethod
    def _from_protobuf(cls, proto_object, ir):
        # type: (typing.Any, typing.Optional["IR"]) -> Node
        """Deserialize a Node from Protobuf.

        Performs a cache lookup for the object's UUID in the cache, calling the
        class' _decode_protobuf constructor if cannot find it.
        """

        uuid = UUID(bytes=proto_object.uuid)
        node = None
        if ir is not None:
            node = ir.get_by_uuid(uuid)
        if node is None:
            node = cls._decode_protobuf(proto_object, uuid, ir)
        return node

    def _to_protobuf(self):
        # type: () -> typing.Any
        """Get a Protobuf representation of ``self``.
        Must be overridden by subclasses.
        """

        raise NotImplementedError

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        """Check: is ``self`` structurally equal to ``other``?

        This method should be used only when deep structural equality checks
        are actually needed, and not for all equality checks. Typically the
        default implmentation of __eq__, which checks pointer equality, is
        sufficient; Nodes are cached such that references to two Nodes with
        the same UUID refer to the same exact object. Use this method when
        you have manually constructed Nodes that may share the same UUID
        despite being different objects, and you need to check for structural
        equality.
        """

        raise NotImplementedError



import typing
from io import BytesIO
from uuid import UUID

from .node import Node
from .proto import AuxData_pb2
from .serialization import Serialization
from .util import DictLike


class AuxData:
    """AuxData objects can be attached to the :class:`gtirb.IR` or individual
    :class:`gtirb.Module` s to store additional client-specific data in a
    portable way.

    AuxData represents a portable, language-independent manner of encoding
    rich data. To do this, all data is stored on disk as a series of bytes
    with a string describing the format of the data, called a *type name*. See
    :mod:`gtirb.serialization` for the list of all default types. Types may
    also be parameterized; for example, ``mapping<string,UUID>`` is a ``dict``
    from ``str`` objects to ``UUID`` objects. All ``AuxData`` requires
    a valid type name in order to be serialized.

    :ivar ~.data: The value stored in this AuxData.
    :ivar ~.type_name: A string describing the type of ``data``.
        Used to determine the proper codec for serializing this AuxData.
    """

    serializer = Serialization()  # type: typing.ClassVar[Serialization]
    """This is a :class:`gtirb.Serialization` instance, used to
    encode and decode ``data`` fields of all ``AuxData``. See
    :mod:`gtirb.serialization` for details.
    """

    def __init__(self, data, type_name):
        # type: (typing.Any, str) -> None
        """
        :param data: The value stored in this AuxData.
        :param type_name: A string describing the type of ``data``.
            Used to determine the proper codec for serializing this AuxData.
        """

        self.data = data  # type: typing.Any
        self.type_name = type_name  # type: str

    @classmethod
    def _from_protobuf(cls, aux_data, ir):
        # type: (AuxData_pb2.AuxData, typing.Optional["IR"]) -> AuxData
        """Deserialize AuxData from Protobuf.

        :param aux_data: The Protobuf AuxData object.
        """

        data = AuxData.serializer.decode(
            BytesIO(aux_data.data), aux_data.type_name, ir.get_by_uuid
        )
        return cls(data=data, type_name=aux_data.type_name)

    def _to_protobuf(self):
        # type: () -> AuxData_pb2.AuxData
        """Get a Protobuf representation of the AuxData."""

        out_bytes_array = BytesIO()
        AuxData.serializer.encode(out_bytes_array, self.data, self.type_name)
        proto_auxdata = AuxData_pb2.AuxData()
        proto_auxdata.type_name = self.type_name
        proto_auxdata.data = out_bytes_array.getvalue()
        return proto_auxdata

    def __repr__(self):
        # type: () -> str
        return (
            "AuxData("
            "type_name={type_name!r}, "
            "data={data!r}, "
            ")".format(**self.__dict__)
        )


class AuxDataContainer(Node):
    """The base class for anything that holds AuxData tables; that is,
    :class:`gtirb.IR` and :class:`gtirb.Module`.

    :ivar ~.aux_data: The auxiliary data associated
            with the object, as a mapping from names to
            :class:`gtirb.AuxData`.
    """

    def __init__(
        self,
        aux_data={},  # type: DictLike[str, AuxData]
        uuid=None,  # type: typing.Optional[UUID]
    ):
        # type: (...) -> None
        """
        :param aux_data: The initial auxiliary data to be associated
            with the object, as a mapping from names to
            :class:`gtirb.AuxData`. Defaults to an empty :class:`dict`.
        :param uuid: the UUID of this ``AuxDataContainer``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """
        super().__init__(uuid)
        self.aux_data = dict(aux_data)  # type: typing.Dict[str, AuxData]

    @classmethod
    def _read_protobuf_aux_data(cls, proto_container, ir):
        # type: (typing.Any,typing.Optional["IR"]) -> typing.Dict[str, AuxData]
        """
        Instead of the overrided _decode_protobuf, this method requires the
        Protobuf message to read from. AuxDataContainers need to call this
        method in their own _decode_protobuf overrides.

        :param proto_container: A Protobuf message with a field called
            ``aux_data``.
        """
        return {
            key: AuxData._from_protobuf(val, ir)
            for key, val in proto_container.aux_data.items()
        }

    def _write_protobuf_aux_data(self, proto_container):
        # type: (typing.Any) -> None
        """
        Instead of the overrided _to_protobuf, this method requires the
        Protobuf message to write into. AuxDataContainers need to call this
        method in their own _to_protobuf overrides.

        :param proto_container: A Protobuf message with a field called
            ``aux_data``.
        """
        for k, v in self.aux_data.items():
            proto_container.aux_data[k].CopyFrom(v._to_protobuf())

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        """This overrides :func:`gtirb.Node.deep_eq` to check for
        AuxData equality.

        Because the values stored by AuxData are not necessarily
        amenable to deep checking, the auxiliary data dictionaries
        stored for ``self`` and ``other`` are not deeply checked. Instead,
        they are considered to be equal if their sets of keys are equal.
        """

        if not isinstance(other, AuxDataContainer):
            return False
        if (
            self.uuid != other.uuid
            or self.aux_data.keys() != other.aux_data.keys()
        ):
            return False
        return True



import typing
from uuid import UUID

from .block import ByteBlock, CodeBlock, DataBlock
from .node import Node
from .proto import ByteInterval_pb2, SymbolicExpression_pb2
from .symbolicexpression import (
    SymAddrAddr,
    SymAddrConst,
    SymbolicExpression,
    SymStackConst,
)
from .util import DictLike, SetWrapper, get_desired_range, nodes_at, nodes_on

SymbolicExpressionElement = typing.Tuple[
    "ByteInterval", int, SymbolicExpression
]


class ByteInterval(Node):
    """A contiguous region of bytes in a binary.

    A ByteInterval defines a relative ordering for a group of
    :class:`ByteBlock`\\s, optionally at a fixed address in memory. It also
    stores the bytes associated with these blocks.

    If two blocks are in two different ByteIntervals, then it should be
    considered safe (that is, preserving of program semantics) to move one
    block relative to the other in memory. If two blocks are in the same
    ByteInterval, then it should be considered unknown if moving the two blocks
    relative to one another in memory is a safe operation.

    :ivar ~.address: The fixed address of this interval, if present. If this
        field is present, it may indicate the original address at which this
        interval was located at in memory, or it may indicate that this block's
        address is fixed and must not be changed. If this field is not present,
        it indicates that the interval is free to be moved around in memory
        while preserving program semantics.
    :ivar ~.size: The size of this interval in bytes. If this number is greater
        than ``initialized_size``, this indicates that the high addresses taken
        up by this interval consist of uninitialized bytes. This often occurs
        in BSS sections, where data is zero-initialized rather than stored as
        zeroes in the binary.
    :ivar ~.contents: The bytes stored in this interval.
    :ivar ~.blocks: A set of all :class:`ByteBlock`\\s in this interval.
    :ivar ~.symbolic_expressions: A mapping, from offset in the interval, to a
        :class:`SymbolicExpression` in the interval.
    """

    class _BlockSet(SetWrapper):
        def __init__(self, node, *args):
            self._node = node  # type: ByteInterval
            super().__init__(*args)

        def add(self, v):
            if v._byte_interval is not None:
                v._byte_interval.blocks.discard(v)
            v._byte_interval = self._node
            if self._node.ir is not None:
                v._add_to_uuid_cache(self._node.ir._local_uuid_cache)
            return super().add(v)

        def discard(self, v):
            if v not in self:
                return
            v._byte_interval = None
            if self._node.ir is not None:
                v._remove_from_uuid_cache(self._node.ir._local_uuid_cache)
            return super().discard(v)

    def __init__(
        self,
        *,
        address=None,  # type: typing.Optional[int]
        size=None,  # type: typing.Optional[int]
        initialized_size=None,  # type: typing.Optional[int]
        contents=b"",  # type: typing.ByteString
        blocks=(),  # type: typing.Iterable[ByteBlock]
        symbolic_expressions={},  # type: DictLike[int, SymbolicExpression]
        uuid=None  # type: typing.Optional[UUID]
    ):
        """
        :param address: The fixed address of this interval, if present.
        :param size: The size of this interval in bytes.
        :param initialized_size: The number of initialized bytes in this
            interval.
        :param contents: The bytes stored in this interval.
        :param blocks: A set of all :class:`ByteBlock`\\s in this interval.
        :param symbolic_expressions: A mapping, from offset in the interval, to
            a :class:`SymbolicExpression` in the interval.
        :param uuid: The UUID of this ``ByteInterval``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """

        if size is None:
            size = len(contents)
        if initialized_size is None:
            initialized_size = len(contents)
        if initialized_size > size:
            raise ValueError("initialized_size must be <= size!")

        super().__init__(uuid=uuid)
        self._section = None  # type: typing.Optional["Section"]
        self.address = address  # type: typing.Optional[int]
        self.size = size  # type: int
        self.contents = bytearray(contents)  # type: bytearray
        self.initialized_size = initialized_size
        self.blocks = ByteInterval._BlockSet(
            self, blocks
        )  # type: typing.Set[ByteBlock]
        self.symbolic_expressions = dict(
            symbolic_expressions
        )  # type: typing.Dict[int, SymbolicExpression]
        self._proto_interval = (
            None
        )  # type: typing.Optional[ByteInterval_pb2.ByteInterval]

    @property
    def initialized_size(self):
        # type: () -> int
        """The number of initialized bytes in this interval.

        Not all bytes in this interval may correspond to bytes physically
        stored in the underlying file format. This can occur, for example, in
        BSS sections, which are zero-initialized at loadtime, but these zeroes
        are not stored in the file itself. If this number is smaller than
        ``size``, this indicates that any bytes past this number are
        unitialized bytes with values determined at loadtime. As such, all
        bytes past this number in this interval's byte vector are truncated
        when saving to file.
        """

        return len(self.contents)

    @initialized_size.setter
    def initialized_size(self, value):
        # type: (int) -> None
        if value > len(self.contents):
            self.contents += b"\0" * (value - len(self.contents))
        elif value < len(self.contents):
            self.contents = self.contents[:value]

    @classmethod
    def _decode_protobuf(
        cls,
        proto_interval,  # type: ByteInterval_pb2.ByteInterval
        uuid,  # type: UUID
        ir,  # type: typing.Optional["IR"]
    ):
        # type: (...) -> ByteInterval

        def decode_block(proto_block):
            if proto_block.HasField("code"):
                block = CodeBlock._from_protobuf(proto_block.code, ir)
            elif proto_block.HasField("data"):
                block = DataBlock._from_protobuf(proto_block.data, ir)
            else:
                raise TypeError(
                    "Unknown type inside proto block: %s"
                    % proto_block.WhichOneof("value")
                )

            block.offset = proto_block.offset
            return block

        # we do not decode symbolic expressions yet, because symbols have
        # not yet been decoded at this point.
        result = cls(
            address=proto_interval.address
            if proto_interval.has_address
            else None,
            size=proto_interval.size,
            contents=proto_interval.contents,
            uuid=uuid,
        )
        result._add_to_uuid_cache(ir._local_uuid_cache)
        result.blocks.update(decode_block(b) for b in proto_interval.blocks)
        # We store the interval and IR here so we can use it later, when
        # _decode_symbolic_expressions is called.
        result._proto_interval = proto_interval
        # Return the new BI.
        return result

    def _decode_symbolic_expressions(self, ir):
        """Called by modules after symbols are decoded, but before the module
        is done decoding.
        """

        def decode_symbolic_expression(proto_expr):
            if proto_expr.HasField("stack_const"):
                return SymStackConst._from_protobuf(
                    proto_expr.stack_const, ir.get_by_uuid
                )
            elif proto_expr.HasField("addr_const"):
                return SymAddrConst._from_protobuf(
                    proto_expr.addr_const, ir.get_by_uuid
                )
            elif proto_expr.HasField("addr_addr"):
                return SymAddrAddr._from_protobuf(
                    proto_expr.addr_addr, ir.get_by_uuid
                )
            else:
                raise TypeError(
                    "Unknown type inside proto sym expr: %s"
                    % proto_expr.WhichOneof("value")
                )

        self.symbolic_expressions = {
            i: decode_symbolic_expression(v)
            for i, v in self._proto_interval.symbolic_expressions.items()
        }

        del self._proto_interval

    def _to_protobuf(self):
        # type: () -> ByteInterval_pb2.ByteInterval
        proto_interval = ByteInterval_pb2.ByteInterval()

        proto_interval.uuid = self.uuid.bytes
        if self.address is None:
            proto_interval.has_address = False
        else:
            proto_interval.has_address = True
            proto_interval.address = self.address
        proto_interval.size = self.size
        proto_interval.contents = bytes(self.contents)

        # Cannot insert blocks using proto_interval.blocks.append() in a loop
        # because append() isn't supported in older versions of protobuf. Use a
        # comprehension and extend() instead.

        def to_proto_block(block):
            proto_block = ByteInterval_pb2.Block()
            proto_block.offset = block.offset
            if isinstance(block, CodeBlock):
                proto_block.code.CopyFrom(block._to_protobuf())
            elif isinstance(block, DataBlock):
                proto_block.data.CopyFrom(block._to_protobuf())
            else:
                raise TypeError(
                    "Unknown block type in interval: %s" % type(block)
                )
            return proto_block

        proto_interval.blocks.extend(to_proto_block(b) for b in self.blocks)

        for k, v in self.symbolic_expressions.items():
            sym_exp = SymbolicExpression_pb2.SymbolicExpression()
            if isinstance(v, SymStackConst):
                sym_exp.stack_const.CopyFrom(v._to_protobuf())
            elif isinstance(v, SymAddrConst):
                sym_exp.addr_const.CopyFrom(v._to_protobuf())
            elif isinstance(v, SymAddrAddr):
                sym_exp.addr_addr.CopyFrom(v._to_protobuf())
            else:
                raise ValueError(
                    "Expected sym expr type in interval: %s" % type(v)
                )
            proto_interval.symbolic_expressions[k].CopyFrom(sym_exp)

        return proto_interval

    @property
    def section(self):
        # type: () -> typing.Optional["Section"]
        """The :class:`Section` this interval belongs to."""

        return self._section

    @section.setter
    def section(self, value):
        # type: (typing.Optional["Section"]) -> None
        if self._section is not None:
            self._section.byte_intervals.discard(self)
        if value is not None:
            value.byte_intervals.add(self)

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, ByteInterval):
            return False
        return (
            self.uuid == other.uuid
            and self.address == other.address
            and self.contents == other.contents
            and self.size == other.size
            and len(self.blocks) == len(other.blocks)
            and all(
                self_node.deep_eq(other_node)
                for self_node, other_node in zip(
                    sorted(self.blocks, key=lambda b: b.uuid),
                    sorted(other.blocks, key=lambda b: b.uuid),
                )
            )
            and len(self.symbolic_expressions)
            == len(other.symbolic_expressions)
            and all(
                self_kv[0] == other_kv[0] and self_kv[1].deep_eq(other_kv[1])
                for self_kv, other_kv in zip(
                    sorted(
                        self.symbolic_expressions.items(), key=lambda kv: kv[0]
                    ),
                    sorted(
                        other.symbolic_expressions.items(),
                        key=lambda kv: kv[0],
                    ),
                )
            )
        )

    def __repr__(self):
        # type: () -> str
        return (
            "ByteInterval("
            "uuid={uuid!r}, "
            "address={address}, "
            "size={size}, "
            "contents={contents!r}, "
            "blocks={blocks!r}, "
            "symbolic_expressions={symbolic_expressions!r}, "
            ")".format(**self.__dict__)
        )

    def byte_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
        """Finds all the byte blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.byte_blocks, addrs)

    def byte_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
        """Finds all the byte blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.byte_blocks, addrs)

    def code_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
        """Finds all the code blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.code_blocks, addrs)

    def code_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
        """Finds all the code blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.code_blocks, addrs)

    def data_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
        """Finds all the data blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.data_blocks, addrs)

    def data_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
        """Finds all the data blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.data_blocks, addrs)

    def symbolic_expressions_at(
        self, addrs  # type: typing.Union[int, range]
    ):
        # type: (...) -> typing.Iterable[SymbolicExpressionElement]
        """Finds all the symbolic expressions that begin at an address or
        range of addresses.

        :param addrs: Either a ``range`` object or a single address.
        :returns: Yields ``(interval, offset, symexpr)`` tuples for every
            symbolic expression in the range.
        """

        if self.address is None:
            return

        addrs = get_desired_range(addrs)
        for i, v in self.symbolic_expressions.items():
            if self.address + i in addrs:
                yield (self, i, v)

    def _add_to_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is added."""

        cache[self.uuid] = self
        for block in self.blocks:
            block._add_to_uuid_cache(cache)

    def _remove_from_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is removed."""

        del cache[self.uuid]
        for block in self.blocks:
            block._remove_from_uuid_cache(cache)

    @property
    def module(self):
        # type: () -> "Module"
        """Get the module this node ultimately belongs to."""
        if self.section is None:
            return None
        return self.section.module

    @property
    def ir(self):
        # type: () -> "IR"
        """Get the IR this node ultimately belongs to."""
        if self.module is None:
            return None
        return self.module.ir



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/Section.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gtirb.proto import ByteInterval_pb2 as gtirb_dot_proto_dot_ByteInterval__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/Section.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_options=_b('\n\032com.grammatech.gtirb.proto'),
  serialized_pb=_b('\n\x19gtirb/proto/Section.proto\x12\x0bgtirb.proto\x1a\x1egtirb/proto/ByteInterval.proto\"\xa4\x01\n\x07Section\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x31\n\x0e\x62yte_intervals\x18\x05 \x03(\x0b\x32\x19.gtirb.proto.ByteInterval\x12/\n\rsection_flags\x18\x06 \x03(\x0e\x32\x18.gtirb.proto.SectionFlagJ\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05R\x07\x61\x64\x64ressR\x04size*~\n\x0bSectionFlag\x12\x15\n\x11Section_Undefined\x10\x00\x12\x0c\n\x08Readable\x10\x01\x12\x0c\n\x08Writable\x10\x02\x12\x0e\n\nExecutable\x10\x03\x12\n\n\x06Loaded\x10\x04\x12\x0f\n\x0bInitialized\x10\x05\x12\x0f\n\x0bThreadLocal\x10\x06\x42\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
  ,
  dependencies=[gtirb_dot_proto_dot_ByteInterval__pb2.DESCRIPTOR,])

_SECTIONFLAG = _descriptor.EnumDescriptor(
  name='SectionFlag',
  full_name='gtirb.proto.SectionFlag',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Section_Undefined', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Readable', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Writable', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Executable', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Loaded', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Initialized', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ThreadLocal', index=6, number=6,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=241,
  serialized_end=367,
)
_sym_db.RegisterEnumDescriptor(_SECTIONFLAG)

SectionFlag = enum_type_wrapper.EnumTypeWrapper(_SECTIONFLAG)
Section_Undefined = 0
Readable = 1
Writable = 2
Executable = 3
Loaded = 4
Initialized = 5
ThreadLocal = 6



_SECTION = _descriptor.Descriptor(
  name='Section',
  full_name='gtirb.proto.Section',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='gtirb.proto.Section.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='gtirb.proto.Section.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='byte_intervals', full_name='gtirb.proto.Section.byte_intervals', index=2,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='section_flags', full_name='gtirb.proto.Section.section_flags', index=3,
      number=6, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=239,
)

_SECTION.fields_by_name['byte_intervals'].message_type = gtirb_dot_proto_dot_ByteInterval__pb2._BYTEINTERVAL
_SECTION.fields_by_name['section_flags'].enum_type = _SECTIONFLAG
DESCRIPTOR.message_types_by_name['Section'] = _SECTION
DESCRIPTOR.enum_types_by_name['SectionFlag'] = _SECTIONFLAG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Section = _reflection.GeneratedProtocolMessageType('Section', (_message.Message,), {
  'DESCRIPTOR' : _SECTION,
  '__module__' : 'gtirb.proto.Section_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.Section)
  })
_sym_db.RegisterMessage(Section)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/CFG.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/CFG.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_options=_b('\n\032com.grammatech.gtirb.proto'),
  serialized_pb=_b('\n\x15gtirb/proto/CFG.proto\x12\x0bgtirb.proto\"U\n\tEdgeLabel\x12\x13\n\x0b\x63onditional\x18\x01 \x01(\x08\x12\x0e\n\x06\x64irect\x18\x02 \x01(\x08\x12#\n\x04type\x18\x03 \x01(\x0e\x32\x15.gtirb.proto.EdgeType\"u\n\x04\x45\x64ge\x12\x13\n\x0bsource_uuid\x18\x01 \x01(\x0c\x12\x13\n\x0btarget_uuid\x18\x02 \x01(\x0c\x12%\n\x05label\x18\x05 \x01(\x0b\x32\x16.gtirb.proto.EdgeLabelJ\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05R\x07\x62ooleanR\x07integer\"G\n\x03\x43\x46G\x12\x10\n\x08vertices\x18\x03 \x03(\x0c\x12 \n\x05\x65\x64ges\x18\x02 \x03(\x0b\x32\x11.gtirb.proto.EdgeJ\x04\x08\x01\x10\x02R\x06\x62locks*t\n\x08\x45\x64geType\x12\x0f\n\x0bType_Branch\x10\x00\x12\r\n\tType_Call\x10\x01\x12\x14\n\x10Type_Fallthrough\x10\x02\x12\x0f\n\x0bType_Return\x10\x03\x12\x10\n\x0cType_Syscall\x10\x04\x12\x0f\n\x0bType_Sysret\x10\x05\x42\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
)

_EDGETYPE = _descriptor.EnumDescriptor(
  name='EdgeType',
  full_name='gtirb.proto.EdgeType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Type_Branch', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Type_Call', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Type_Fallthrough', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Type_Return', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Type_Syscall', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Type_Sysret', index=5, number=5,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=317,
  serialized_end=433,
)
_sym_db.RegisterEnumDescriptor(_EDGETYPE)

EdgeType = enum_type_wrapper.EnumTypeWrapper(_EDGETYPE)
Type_Branch = 0
Type_Call = 1
Type_Fallthrough = 2
Type_Return = 3
Type_Syscall = 4
Type_Sysret = 5



_EDGELABEL = _descriptor.Descriptor(
  name='EdgeLabel',
  full_name='gtirb.proto.EdgeLabel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conditional', full_name='gtirb.proto.EdgeLabel.conditional', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='direct', full_name='gtirb.proto.EdgeLabel.direct', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='gtirb.proto.EdgeLabel.type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38,
  serialized_end=123,
)


_EDGE = _descriptor.Descriptor(
  name='Edge',
  full_name='gtirb.proto.Edge',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source_uuid', full_name='gtirb.proto.Edge.source_uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='target_uuid', full_name='gtirb.proto.Edge.target_uuid', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='label', full_name='gtirb.proto.Edge.label', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=125,
  serialized_end=242,
)


_CFG = _descriptor.Descriptor(
  name='CFG',
  full_name='gtirb.proto.CFG',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='vertices', full_name='gtirb.proto.CFG.vertices', index=0,
      number=3, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='edges', full_name='gtirb.proto.CFG.edges', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=244,
  serialized_end=315,
)

_EDGELABEL.fields_by_name['type'].enum_type = _EDGETYPE
_EDGE.fields_by_name['label'].message_type = _EDGELABEL
_CFG.fields_by_name['edges'].message_type = _EDGE
DESCRIPTOR.message_types_by_name['EdgeLabel'] = _EDGELABEL
DESCRIPTOR.message_types_by_name['Edge'] = _EDGE
DESCRIPTOR.message_types_by_name['CFG'] = _CFG
DESCRIPTOR.enum_types_by_name['EdgeType'] = _EDGETYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EdgeLabel = _reflection.GeneratedProtocolMessageType('EdgeLabel', (_message.Message,), {
  'DESCRIPTOR' : _EDGELABEL,
  '__module__' : 'gtirb.proto.CFG_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.EdgeLabel)
  })
_sym_db.RegisterMessage(EdgeLabel)

Edge = _reflection.GeneratedProtocolMessageType('Edge', (_message.Message,), {
  'DESCRIPTOR' : _EDGE,
  '__module__' : 'gtirb.proto.CFG_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.Edge)
  })
_sym_db.RegisterMessage(Edge)

CFG = _reflection.GeneratedProtocolMessageType('CFG', (_message.Message,), {
  'DESCRIPTOR' : _CFG,
  '__module__' : 'gtirb.proto.CFG_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.CFG)
  })
_sym_db.RegisterMessage(CFG)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import itertools
import typing
from enum import Enum
from uuid import UUID

from .block import ByteBlock, CodeBlock, DataBlock
from .byteinterval import ByteInterval, SymbolicExpressionElement
from .node import Node
from .proto import Section_pb2
from .util import SetWrapper, nodes_at, nodes_on, symbolic_expressions_at


class Section(Node):
    """Represents a named section of the binary.

    Does not directly store the contents of the section, which are
    kept in a :class:`gtirb.ImageByteMap`.

    :ivar ~.name: The section name (E.g. ".text", ".bss", etc).
    :ivar ~.byte_intervals: The :class:`ByteInterval`\\s in this section.
    :ivar ~.flags: The :class:`Section.Flag`\\s this section has.
    """

    class Flag(Enum):
        """A flag representing a known property of a section."""

        Undefined = Section_pb2.SectionFlag.Value("Section_Undefined")
        """This value is defined for Protobuf compatibility. Do not use."""

        Readable = Section_pb2.SectionFlag.Value("Readable")
        """This section can be read from at runtime."""

        Writable = Section_pb2.SectionFlag.Value("Writable")
        """This section can be written to at runtime."""

        Executable = Section_pb2.SectionFlag.Value("Executable")
        """This section contains executable code."""

        Loaded = Section_pb2.SectionFlag.Value("Loaded")
        """This section is present in memory at runtime."""

        Initialized = Section_pb2.SectionFlag.Value("Initialized")
        """This section has bytes allocated to it in the binary file."""

        ThreadLocal = Section_pb2.SectionFlag.Value("ThreadLocal")
        """This section is created in memory once per thread."""

    class _ByteIntervalSet(SetWrapper):
        def __init__(self, node, *args):
            self._node = node  # type: Section
            super().__init__(*args)

        def add(self, v):
            if v._section is not None:
                v._section.byte_intervals.discard(v)
            v._section = self._node
            if self._node.ir is not None:
                v._add_to_uuid_cache(self._node.ir._local_uuid_cache)
            return super().add(v)

        def discard(self, v):
            if v not in self:
                return
            v._section = None
            if self._node.ir is not None:
                v._remove_from_uuid_cache(self._node.ir._local_uuid_cache)
            return super().discard(v)

    def __init__(
        self,
        *,
        name="",  # type: str
        byte_intervals=(),  # type: typing.Iterable[ByteInterval]
        flags=set(),  # type: typing.Iterable[Section.Flag]
        uuid=None  # type: typing.Optional[UUID]
    ):
        """
        :param name: The name of this section.
        :param byte_intervals: The :class:`ByteInterval`\\s in this section.
        :param flags: The :class:`Section.Flag`\\s this section has.
        :param uuid: The UUID of this ``Section``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """

        super().__init__(uuid)
        self._module = None  # type: "Module"
        self.name = name  # type: str
        self.byte_intervals = Section._ByteIntervalSet(
            self, byte_intervals
        )  # type: typing.Set[ByteInterval]
        self.flags = set(flags)  # type: typing.Set[Section.Flag]

    @classmethod
    def _decode_protobuf(cls, proto_section, uuid, ir):
        # type: (Section_pb2.Section, UUID, typing.Optional["IR"]) -> Section
        s = cls(
            name=proto_section.name,
            flags=(Section.Flag(f) for f in proto_section.section_flags),
            uuid=uuid,
        )
        s._add_to_uuid_cache(ir._local_uuid_cache)
        s.byte_intervals.update(
            ByteInterval._from_protobuf(bi, ir)
            for bi in proto_section.byte_intervals
        )
        return s

    def _to_protobuf(self):
        # type: () -> Section_pb2.Section
        """Get a Protobuf representation of ``self``."""

        proto_section = Section_pb2.Section()
        proto_section.uuid = self.uuid.bytes
        proto_section.name = self.name
        proto_section.byte_intervals.extend(
            bi._to_protobuf() for bi in self.byte_intervals
        )
        proto_section.section_flags.extend(f.value for f in self.flags)
        return proto_section

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, Section):
            return False
        return (
            self.uuid == other.uuid
            and self.name == other.name
            and len(self.byte_intervals) == len(other.byte_intervals)
            and all(
                self_node.deep_eq(other_node)
                for self_node, other_node in zip(
                    sorted(self.byte_intervals, key=lambda bi: bi.uuid),
                    sorted(other.byte_intervals, key=lambda bi: bi.uuid),
                )
            )
            and self.flags == other.flags
        )

    def __repr__(self):
        # type: () -> str
        return (
            "Section("
            "uuid={uuid!r}, "
            "name={name!r}, "
            "byte_intervals={byte_intervals!r}, "
            "flags=({pretty_flags}), "
            ")".format(
                pretty_flags=", ".join(
                    "Section." + str(f) for f in self.flags
                ),
                **self.__dict__,
            )
        )

    @property
    def module(self):
        # type: () -> "Module"
        """The :class:`Module` this section belongs to."""

        return self._module

    @module.setter
    def module(self, value):
        # type: ("Module") -> None
        if self._module is not None:
            self._module.sections.discard(self)
        if value is not None:
            value.sections.add(self)

    @property
    def byte_blocks(self):
        # type: () -> typing.Iterator[ByteBlock]
        """The :class:`ByteBlock`\\s in this section."""

        return itertools.chain.from_iterable(
            bi.blocks for bi in self.byte_intervals
        )

    @property
    def code_blocks(self):
        # type: () -> typing.Iterator[CodeBlock]
        """The :class:`CodeBlock`\\s in this section."""

        return (b for b in self.byte_blocks if isinstance(b, CodeBlock))

    @property
    def data_blocks(self):
        # type: () -> typing.Iterator[DataBlock]
        """The :class:`DataBlock`\\s in this section."""

        return (b for b in self.byte_blocks if isinstance(b, DataBlock))

    @property
    def address(self):
        # type: () -> typing.Optional[int]
        """Get the address of this section, if known.

        The address is calculated from the :class:`ByteInterval` objects in
        this section. More specifically, if the address of all byte intervals
        in this section are fixed, then it will return the address of the
        interval lowest in memory. If any one interval does not have an address
        then this will be ``None``, as the address is not calculable in that
        case. Note that a section with no intervals in it has no address or
        size, so it will be ``None`` in that case.
        """

        lowest = None
        for bi in self.byte_intervals:
            if bi.address is None:
                return None
            if lowest is None or bi.address < lowest:
                lowest = bi.address
        return lowest

    @property
    def size(self):
        # type: () -> typing.Optional[int]
        """Get the size of this section, if known.

        The address is calculated from the :class:`ByteInterval` objects in
        this section. More specifically, if the address of all byte intervals
        in this section are fixed, then it will return the difference between
        the lowest and highest address among the intervals. If any one interval
        does not have an address, then this will be ``None``, as the size is
        not calculable in that case. Note that a section with no intervals in
        it has no address or size, so it will be ``None`` in that case.
        """

        lowest = None
        highest = None
        for bi in self.byte_intervals:
            if bi.address is None:
                return None
            if lowest is None or bi.address < lowest:
                lowest = bi.address
            if highest is None or bi.address + bi.size > highest:
                highest = bi.address + bi.size
        if lowest is None or highest is None:
            return None
        return highest - lowest

    def byte_intervals_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
        """Finds all the byte intervals that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.byte_intervals, addrs)

    def byte_intervals_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
        """Finds all the byte intervals that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.byte_intervals, addrs)

    def byte_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
        """Finds all the byte blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.byte_blocks, addrs)

    def byte_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
        """Finds all the byte blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.byte_blocks, addrs)

    def code_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
        """Finds all the code blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.code_blocks, addrs)

    def code_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
        """Finds all the code blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.code_blocks, addrs)

    def data_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
        """Finds all the data blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.data_blocks, addrs)

    def data_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
        """Finds all the data blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.data_blocks, addrs)

    def symbolic_expressions_at(
        self, addrs  # type: typing.Union[int, range]
    ):
        # type: (...) -> typing.Iterable[SymbolicExpressionElement]
        """Finds all the symbolic expressions that begin at an address or
        range of addresses.

        :param addrs: Either a ``range`` object or a single address.
        :returns: Yields ``(interval, offset, symexpr)`` tuples for every
            symbolic expression in the range.
        """

        return symbolic_expressions_at(self.byte_intervals, addrs)

    def _add_to_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is added."""

        cache[self.uuid] = self
        for bi in self.byte_intervals:
            bi._add_to_uuid_cache(cache)

    def _remove_from_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is removed."""

        del cache[self.uuid]
        for bi in self.byte_intervals:
            bi._remove_from_uuid_cache(cache)

    @property
    def ir(self):
        # type: () -> "IR"
        """Get the IR this node ultimately belongs to."""
        if self.module is None:
            return None
        return self.module.ir



import glob
import setuptools
import sys
import unittest

# if version of setuptools is 20.8.1 or below, it doesn't support env specs
if int(setuptools.__version__.split(".")[0]) < 21:
    if sys.version_info[:3] < (3, 5, 0):
        typing_dep = ["typing"]
    else:
        typing_dep = []
else:
    typing_dep = ["typing ; python_version<'3.5.0'"]


def gtirb_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover("tests", pattern="test_*.py")
    return test_suite


if __name__ == "__main__":
    setuptools.setup(
        name="gtirb",
        version=(
            "1"
            ".6"
            ".0"
        ),
        author="abhaskar",
        author_email="abhaskar@grammatech.com",
        description="The gtirb package",
        packages=setuptools.find_packages(),
        test_suite="setup.gtirb_test_suite",
        install_requires=["protobuf"] + typing_dep,
        classifiers=["Programming Language :: Python :: 3"],
    )



from uuid import UUID

from .proto import Offset_pb2


class Offset:
    """
    An Offset describes a location inside a :class:`gtirb.CodeBlock`
    or :class:`gtirb.DataBlock`.

    :ivar ~.element_id: The UUID of a :class:`gtirb.ByteBlock`
            containing the location of interest.
    :ivar ~.displacement: The offset inside the Node to point to.
    """

    def __init__(self, element_id, displacement):
        # type: (UUID,int) -> None
        """
        :param element_id: The UUID of a :class:`gtirb.ByteBlock`
            containing the location of interest.
        :param displacement: The offset inside the Node to point to.
        """

        self.element_id = element_id  # type: UUID
        self.displacement = displacement  # type: int

    @classmethod
    def _from_protobuf(cls, offset):
        # type: (Offset_pb2.Offset) -> Offset
        """Decode a Protobuf object to an offset.

        :param offset: The Protobuf object.
        """

        return cls(UUID(bytes=offset.element_id), offset.displacement)

    def _to_protobuf(self):
        # type: () -> Offset_pb2.Offset
        """Encode this offset into a Protobuf object."""

        proto_offset = Offset_pb2.Offset()
        proto_offset.element_id = self.element_id.bytes
        proto_offset.displacement = self.displacement
        return proto_offset

    def __repr__(self):
        # type: () -> str
        return (
            "Offset("
            "element_id={element_id!r}, "
            "displacement={displacement!r}, "
            ")".format(**self.__dict__)
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/ByteInterval.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gtirb.proto import CodeBlock_pb2 as gtirb_dot_proto_dot_CodeBlock__pb2
from gtirb.proto import DataBlock_pb2 as gtirb_dot_proto_dot_DataBlock__pb2
from gtirb.proto import SymbolicExpression_pb2 as gtirb_dot_proto_dot_SymbolicExpression__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/ByteInterval.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_options=_b('\n\032com.grammatech.gtirb.proto'),
  serialized_pb=_b('\n\x1egtirb/proto/ByteInterval.proto\x12\x0bgtirb.proto\x1a\x1bgtirb/proto/CodeBlock.proto\x1a\x1bgtirb/proto/DataBlock.proto\x1a$gtirb/proto/SymbolicExpression.proto\"p\n\x05\x42lock\x12\x0e\n\x06offset\x18\x01 \x01(\x04\x12&\n\x04\x63ode\x18\x02 \x01(\x0b\x32\x16.gtirb.proto.CodeBlockH\x00\x12&\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x16.gtirb.proto.DataBlockH\x00\x42\x07\n\x05value\"\xb5\x02\n\x0c\x42yteInterval\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\"\n\x06\x62locks\x18\x02 \x03(\x0b\x32\x12.gtirb.proto.Block\x12P\n\x14symbolic_expressions\x18\x03 \x03(\x0b\x32\x32.gtirb.proto.ByteInterval.SymbolicExpressionsEntry\x12\x13\n\x0bhas_address\x18\x04 \x01(\x08\x12\x0f\n\x07\x61\x64\x64ress\x18\x05 \x01(\x04\x12\x0c\n\x04size\x18\x06 \x01(\x04\x12\x10\n\x08\x63ontents\x18\x07 \x01(\x0c\x1a[\n\x18SymbolicExpressionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x04\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x1f.gtirb.proto.SymbolicExpression:\x02\x38\x01\x42\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
  ,
  dependencies=[gtirb_dot_proto_dot_CodeBlock__pb2.DESCRIPTOR,gtirb_dot_proto_dot_DataBlock__pb2.DESCRIPTOR,gtirb_dot_proto_dot_SymbolicExpression__pb2.DESCRIPTOR,])




_BLOCK = _descriptor.Descriptor(
  name='Block',
  full_name='gtirb.proto.Block',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='offset', full_name='gtirb.proto.Block.offset', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='code', full_name='gtirb.proto.Block.code', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='gtirb.proto.Block.data', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='gtirb.proto.Block.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=143,
  serialized_end=255,
)


_BYTEINTERVAL_SYMBOLICEXPRESSIONSENTRY = _descriptor.Descriptor(
  name='SymbolicExpressionsEntry',
  full_name='gtirb.proto.ByteInterval.SymbolicExpressionsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='gtirb.proto.ByteInterval.SymbolicExpressionsEntry.key', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='gtirb.proto.ByteInterval.SymbolicExpressionsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=476,
  serialized_end=567,
)

_BYTEINTERVAL = _descriptor.Descriptor(
  name='ByteInterval',
  full_name='gtirb.proto.ByteInterval',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='gtirb.proto.ByteInterval.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blocks', full_name='gtirb.proto.ByteInterval.blocks', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbolic_expressions', full_name='gtirb.proto.ByteInterval.symbolic_expressions', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='has_address', full_name='gtirb.proto.ByteInterval.has_address', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='address', full_name='gtirb.proto.ByteInterval.address', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size', full_name='gtirb.proto.ByteInterval.size', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contents', full_name='gtirb.proto.ByteInterval.contents', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_BYTEINTERVAL_SYMBOLICEXPRESSIONSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=258,
  serialized_end=567,
)

_BLOCK.fields_by_name['code'].message_type = gtirb_dot_proto_dot_CodeBlock__pb2._CODEBLOCK
_BLOCK.fields_by_name['data'].message_type = gtirb_dot_proto_dot_DataBlock__pb2._DATABLOCK
_BLOCK.oneofs_by_name['value'].fields.append(
  _BLOCK.fields_by_name['code'])
_BLOCK.fields_by_name['code'].containing_oneof = _BLOCK.oneofs_by_name['value']
_BLOCK.oneofs_by_name['value'].fields.append(
  _BLOCK.fields_by_name['data'])
_BLOCK.fields_by_name['data'].containing_oneof = _BLOCK.oneofs_by_name['value']
_BYTEINTERVAL_SYMBOLICEXPRESSIONSENTRY.fields_by_name['value'].message_type = gtirb_dot_proto_dot_SymbolicExpression__pb2._SYMBOLICEXPRESSION
_BYTEINTERVAL_SYMBOLICEXPRESSIONSENTRY.containing_type = _BYTEINTERVAL
_BYTEINTERVAL.fields_by_name['blocks'].message_type = _BLOCK
_BYTEINTERVAL.fields_by_name['symbolic_expressions'].message_type = _BYTEINTERVAL_SYMBOLICEXPRESSIONSENTRY
DESCRIPTOR.message_types_by_name['Block'] = _BLOCK
DESCRIPTOR.message_types_by_name['ByteInterval'] = _BYTEINTERVAL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Block = _reflection.GeneratedProtocolMessageType('Block', (_message.Message,), {
  'DESCRIPTOR' : _BLOCK,
  '__module__' : 'gtirb.proto.ByteInterval_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.Block)
  })
_sym_db.RegisterMessage(Block)

ByteInterval = _reflection.GeneratedProtocolMessageType('ByteInterval', (_message.Message,), {

  'SymbolicExpressionsEntry' : _reflection.GeneratedProtocolMessageType('SymbolicExpressionsEntry', (_message.Message,), {
    'DESCRIPTOR' : _BYTEINTERVAL_SYMBOLICEXPRESSIONSENTRY,
    '__module__' : 'gtirb.proto.ByteInterval_pb2'
    # @@protoc_insertion_point(class_scope:gtirb.proto.ByteInterval.SymbolicExpressionsEntry)
    })
  ,
  'DESCRIPTOR' : _BYTEINTERVAL,
  '__module__' : 'gtirb.proto.ByteInterval_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.ByteInterval)
  })
_sym_db.RegisterMessage(ByteInterval)
_sym_db.RegisterMessage(ByteInterval.SymbolicExpressionsEntry)


DESCRIPTOR._options = None
_BYTEINTERVAL_SYMBOLICEXPRESSIONSENTRY._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/Symbol.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/Symbol.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_options=_b('\n\032com.grammatech.gtirb.proto'),
  serialized_pb=_b('\n\x18gtirb/proto/Symbol.proto\x12\x0bgtirb.proto\"\x86\x01\n\x06Symbol\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x0f\n\x05value\x18\x02 \x01(\x04H\x00\x12\x17\n\rreferent_uuid\x18\x05 \x01(\x0cH\x00\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0e\n\x06\x61t_end\x18\x06 \x01(\x08\x42\x12\n\x10optional_payloadJ\x04\x08\x04\x10\x05R\x0cstorage_kindB\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
)




_SYMBOL = _descriptor.Descriptor(
  name='Symbol',
  full_name='gtirb.proto.Symbol',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='gtirb.proto.Symbol.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='gtirb.proto.Symbol.value', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='referent_uuid', full_name='gtirb.proto.Symbol.referent_uuid', index=2,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='gtirb.proto.Symbol.name', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='at_end', full_name='gtirb.proto.Symbol.at_end', index=4,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='optional_payload', full_name='gtirb.proto.Symbol.optional_payload',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=42,
  serialized_end=176,
)

_SYMBOL.oneofs_by_name['optional_payload'].fields.append(
  _SYMBOL.fields_by_name['value'])
_SYMBOL.fields_by_name['value'].containing_oneof = _SYMBOL.oneofs_by_name['optional_payload']
_SYMBOL.oneofs_by_name['optional_payload'].fields.append(
  _SYMBOL.fields_by_name['referent_uuid'])
_SYMBOL.fields_by_name['referent_uuid'].containing_oneof = _SYMBOL.oneofs_by_name['optional_payload']
DESCRIPTOR.message_types_by_name['Symbol'] = _SYMBOL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Symbol = _reflection.GeneratedProtocolMessageType('Symbol', (_message.Message,), {
  'DESCRIPTOR' : _SYMBOL,
  '__module__' : 'gtirb.proto.Symbol_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.Symbol)
  })
_sym_db.RegisterMessage(Symbol)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)
